#include "FresnelSurface.h"
#include "FresnelOutput_def.h"
#include "LensUtils.h"
#include "CylindricSurface.h"

#include <CLHEP/Matrix/Matrix.h>
using CLHEP::HepMatrix;

#include <G4GeometryTolerance.hh>
#include <globals.hh>

#include <float.h>
#include <stdarg.h>

using namespace G4FresnelLens;
using std::swap;
using std::min;
using std::max;
using std::vector;

//__________________________________________________________________________________
FresnelSurface::FresnelSurface(const std::vector<SSurface*>& surf,
                               ParametricSurface* downsurf,
                               ParametricSurface* upsurf):
SSurface(),
fUpLimit(upsurf),
fDownLimit(downsurf),
theToothEdges(0)
{
    FillArray(surf);
}

//__________________________________________________________________________________
FresnelSurface::~FresnelSurface(){
    if (theToothEdges) delete [] theToothEdges;
    if ( fUpLimit ) delete fUpLimit;
    if ( fDownLimit ) delete fDownLimit;

    for (unsigned i(0); i<fSurfaces.size(); i++){
        delete fSurfaces[i];
    }
    fSurfaces.clear();

    for (unsigned i(0); i<fSurfacesIns.size(); i++){
        if (i%2) delete fSurfacesIns[i];
    }
    fSurfacesIns.clear();
}

//__________________________________________________________________________________
int FresnelSurface::FindTooth( double rho ) { 
    return BinarySearch(theToothEdges, rho, 0, theToothN); 
} 

//__________________________________________________________________________________
double FresnelSurface::PointOnSurf(double rho) { 
    int n_tooth = FindTooth( rho );
    if( n_tooth<0 ) return kInfinity;
    return fSurfacesIns[ n_tooth*2 ]->PointOnSurf( rho );   
}

//__________________________________________________________________________________
void FresnelSurface::FillArray(const std::vector<SSurface*>& surf){
    #ifdef DEBUG
      theDebugSwitch=false;
    #endif
    theLastPerp=-1.;
    theLastInside=(EInside)0;
    if (theToothEdges) delete [] theToothEdges;

    fR=surf.back()->GetR2();
    fUpLimit->SetR(fR);
    fUpLimit->ShiftSurfaceTo(surf, false);
    fDownLimit->SetR(fR);
    fDownLimit->ShiftSurfaceTo(surf, true);

    fCylinder.SetR(fR);
    fCylinder.SetZ(fDownLimit->GetMinZ(), fUpLimit->GetMaxZ());

    theToothN=surf.size();
    theToothEdges = new double[theToothN+1];
    theToothEdges[0]=0.;

    for (unsigned i(0); i<surf.size()-1; i++){
        double r=surf[i]->GetR();
        theToothEdges[i+1]=r;
        fSurfaces.push_back(surf[i]);
        fSurfaces.push_back(
            new CylindricSurface(surf[i]->GetZ2(),
                                 surf[i+1]->GetZ1(), r));
        ((CylindricSurface*)fSurfaces.back())->SetZ(fDownLimit->PointOnSurf(r), fUpLimit->PointOnSurf(r));
        fSurfacesIns.push_back(surf[i]);
        fSurfacesIns.push_back(
            new CylindricSurface(surf[i]->GetZ2(),
                                 surf[i+1]->GetZ1(), r));
    }
    fSurfaces.push_back(surf[surf.size()-1]);
    fSurfacesIns.push_back(surf[surf.size()-1]);
    theToothEdges[surf.size()]=surf.back()->GetR();

//     #define PRINT
    #ifdef PRINT
    fUpLimit->print();
    fDownLimit->print();
    for (unsigned int i(0); i<fSurfaces.size(); i++){
        printf("%5u  ", i); fSurfaces[i]->print();
        if (i%2) {printf("%5u* ", i); fSurfacesIns[i]->print();}
    }
    #endif
}

//__________________________________________________________________________________
#ifdef USE_ROOT
#include <TGraph.h>
#include <TCanvas.h>
#include <TString.h>
#include <TFile.h>
void FresnelSurface::draw(const char* file, const char* canvasfile, double shift, const char* prefix, const char* option){
    TString file1=file;
    TFile* f = new TFile(file1, option);
    TGraph surf, ulim, dlim;
    ulim.SetLineColor(kGreen);
    dlim.SetLineColor(kRed);

    vector<SSurface*>::iterator it=fSurfaces.begin();
    for ( ; it<fSurfaces.end(); it+=2 ){
        int i(0);
        double rho,z;
        while((*it)->Point(i, 0., rho, z)) {
            surf.SetPoint(surf.GetN(), rho, z+shift);
            i++;
        }
    }

    double step=0.01, rho, z;
    int i(0);
    while( fUpLimit->Point(i, step, rho, z) ){
        ulim.SetPoint(surf.GetN(), rho, z+shift);
        i++;
    }
    i=0;
    while( fDownLimit->Point(i, step, rho, z) ){
        ulim.SetPoint(surf.GetN(), rho, z+shift);
        i++;
    }
    surf.Write( TString(prefix)+"surf" );
    ulim.Write( TString(prefix)+"up" );
    dlim.Write( TString(prefix)+"down" );
    f->Close();

    if (canvasfile) {
        TCanvas c("surface", "surface");
        surf.Draw("AL");
        ulim.Draw("L");
        dlim.Draw("L");
        c.Print(canvasfile);
    }
}
#endif

//__________________________________________________________________________________
EInside FresnelSurface::Inside(double rho, double z, FSIntersectedSegment *intSeg){
    #ifdef DEBUG
      if ( theDebugSwitch ){
          printf("FresnelSurface::Inside (%.10f, %.10f)\n", rho, z);
      }
    #endif
    if ( fabs(theLastPerp-rho)<kHalfTolerance && fabs(theLastZ-rho)<kHalfTolerance ) {
         #ifdef DEBUG
           if ( theDebugSwitch ){
               printf( "Return last saved state (%.10f, %.10f): %s\n", rho, z,
                       theLastInside==kSurface?"surface":(theLastInside==kInside?"inside":"outside"));
           }
         #endif
         if ( intSeg && theLastInside==kSurface ) *intSeg=theLastIntSeg;
         return theLastInside;
    }
    theLastPerp=rho;
    theLastZ=z;

    #ifdef DEBUG
      if ( theDebugSwitch ){
          printf("Check top/bottom coordinates: b(%.10f), z(%.10f), t(%.10f)\n", this->GetBottomZ(rho), z, this->GetUpperZ(rho) );
      }
    #endif
    if ( z<GetBottomZ(rho)) return theLastInside=kInside;
    if ( z>GetUpperZ(rho) ) return theLastInside=kOutside;

    #ifdef DEBUG
      if ( theDebugSwitch ){
          printf("Find tooth\n");
      }
    #endif
    int n_tooth = FindTooth( rho );
    if( n_tooth<0 ) {
        printf("Warning. Tooth not found. Shouldn't happen (rho=%g of [%g, %g]).",
               rho, theToothEdges[0], theToothEdges[theToothN]);
        return theLastInside=kOutside;
    }

    vector<SSurface*>::const_iterator it=fSurfacesIns.begin()+2*n_tooth;
    if ( (n_tooth                 && (*(it-1))->Inside(rho, z, intSeg)==kSurface) || //left
         (it+1<fSurfacesIns.end() && (*(it+1))->Inside(rho, z, intSeg)==kSurface)    //right
       ) {
           if (intSeg) theLastIntSeg=*intSeg;
           return theLastInside=kSurface;
       }
    theLastInside=(*it)->Inside(rho, z, intSeg);                              //surface
    if ( intSeg && theLastInside==kSurface ) theLastIntSeg=*intSeg;
    return theLastInside;
}

//__________________________________________________________________________________
double FresnelSurface::FindNearestIntersection(const G4ThreeVector& p, const G4ThreeVector& v, FSIntersectedSegment *intSeg){
    //
    // find nearest intersection with fresnel surface
    //
    #ifdef DEBUG
      if (theDebugSwitch) {
          info_<<"FindeNearestIntersection started"<<endline_;
      }
    #endif

    // initialize tracking variables
    G4ThreeVector pos=p;
    double perp=p.perp();
    double shift=0.;

    {// approach to the limiting surface
        #ifdef DEBUG
          if (theDebugSwitch) {
              info_<<"Approaching"<<endline_;
          }
        #endif
        double dist(kInfinity);
        double z_down=this->GetBottomZ(perp);
        if ( pos.z()<z_down ) dist=fDownLimit->DistanceToSurf(pos, v);
        else {
            double z_up=this->GetUpperZ(perp);
            if ( pos.z()>z_up ) {
                dist=fUpLimit->DistanceToSurf(pos, v);
            }
        }
        #ifdef DEBUG
        if (theDebugSwitch) {
            printf("ddd %e z %e zdown %e\n", dist, pos.z(), z_down);
        }
        #endif
        if ( dist==kInfinity ) {
            if ( perp>fR ){
                dist=fCylinder.DistanceToSurf(pos, v);
                if ( dist==kInfinity ) return kInfinity;
                pos+=v*dist;
                shift+=dist;
            }
        }
        else {
            pos+=v*dist;
            shift+=dist;
        }
        perp=pos.perp();
        #ifdef DEBUG
        if (theDebugSwitch) {
            printf("perp=%f shift=%f z=%f\n", perp, shift, pos.z());
        }
        #endif
    }// approach to the limiting surface

    int theCurrentToothIndex=FindTooth( perp );
    if ( theCurrentToothIndex<0 ) return kInfinity;
    vector<SSurface*>::const_iterator it=fSurfaces.begin()+theCurrentToothIndex*2;
    #ifdef DEBUG
    if (theDebugSwitch) {
        printf("i=%i rhotooth=%f\n", theCurrentToothIndex, theToothEdges[theCurrentToothIndex]);
    }
    #endif

    while ( true ){
        if ( intSeg ) intSeg->Reset();
        int direction;
        {// determine next tooth index
            double pv=pos.x()*v.x()+pos.y()*v.y();
            double c2(0.), c2a(0.);
            //printf("%g pv\n", pv);
            if ( pv>=0. ) {
                direction=+1;
            }
            else {
                double ip2=1./pos.perp2();
                double tp=(*it)->GetR1();
                //printf("tp %g\n", tp);
                if ( tp==0. ) direction=+1.;
                else {
                    double r_over_rho=tp*tp*ip2;
                    if ( r_over_rho>1. ) r_over_rho=1.;
                    c2a=r_over_rho-1.;
                    c2=(pv>=0.?1.:-1.)*pv*pv*ip2/v.perp2();
                    //printf("c2 %g c2a %g\n", c2, c2a);
                    if ( c2<-1. ) c2=-1.;
                    direction = c2>=c2a?+1:-1;
                }
            }
        }// determine next tooth index
        vector<SSurface*>::const_iterator next=it+direction;

        FSIntersectedSegment cylseg;
        double dist_cyl = next<fSurfaces.end() ? (*next)->DistanceToSurf(pos, v, &cylseg) : kInfinity;
        double dist_cone=(*it)->DistanceToSurf(pos, v, intSeg);
        if ( dist_cone<dist_cyl ) {
            if ( ( (dist_cyl-dist_cone)<=kTolerance ) ) {
                // skip edge with the thickness less than kTolerance,
                it+=2*direction;
                if ( it>=fSurfaces.end() ) return kInfinity;
                shift+=dist_cyl; pos+=v*dist_cyl; perp=pos.perp();
                continue;
            }
            return shift+dist_cone;
        }
        else if ( dist_cyl==kInfinity ) return kInfinity;

        {//shift to cylinder
            int step=theCurrentToothIndex*2+direction; 
            if ( step<0 || step>=(int)fSurfacesIns.size() ) {
                printf("Trying to step to the unexistent cylinder:\n");
                printf("idx=%i, dir=%i, step=%i, size=%i\n", theCurrentToothIndex, direction, step, (int)fSurfacesIns.size());
                return kInfinity;
            }

            if ( dist_cyl<kHalfTolerance ) dist_cyl=kHalfTolerance;
            shift+=dist_cyl; pos+=v*dist_cyl; perp=pos.perp();

            vector<SSurface*>::const_iterator it1=fSurfacesIns.begin()+step;
            if ( (pos.z()-(*it1)->GetZ1())*(pos.z()-(*it1)->GetZ2())<0 ){
                if ( (*(it1+direction))->TestEdgeTolerance(direction, pos, v)>kTolerance ){
                    if ( intSeg ) *intSeg=cylseg;
                    return shift;
                }
                shift+=kTolerance; pos+=v*kTolerance; perp=pos.perp();
            }
        }//shift to cylinder
        it+=2*direction;
        if ( it>=fSurfaces.end() ) return kInfinity;
    }
}

//______________________________________________________________________________
double* FresnelSurface::CopyTeeth(int& n){
    double* teeth=new double [theToothN+1];
    n=theToothN+1;
    memcpy(teeth, theToothEdges, (theToothN+1)*sizeof(double));
    return teeth;
}
