#include "ShapeOfLens.h"
#include "FresnelSurface.h"
#include "LensUtils.h"
#include "FresnelOutput_def.h"

#include <globals.hh>
#include <G4VoxelLimits.hh>
#include <G4AffineTransform.hh>
#include <G4GeometryTolerance.hh>
#include <G4OpBoundaryProcess.hh>
#include <G4Tubs.hh>
#include <G4EventManager.hh>
#include <G4Track.hh>
#include <Randomize.hh>

using namespace G4FresnelLens;
using namespace CLHEP;
using std::min;
using std::max;
using std::swap;

//#define DEBUG_WNF
//#define DEBUG
//#define DEBUG_SLANTWALLS

///////////////////////////////////////////////////////////////////////////////
//
// constructor - check parameters
//__________________________________________________________________________________
ShapeOfLens::ShapeOfLens(const G4String& pName,
                               G4double pDz,
                               G4double pR,
                               SSurface* upsurf, SSurface* downsurf)
: G4VSolid(pName),
__INHERITCONSTRUCTOR("ShapeOfLens")
fCubicVolume(0.), //fpPolyhedron(0),
fUpSurface(upsurf), fDownSurface(downsurf)
#ifdef SLANTWALLS
,isSlant(false), theSlantSigma(0.)
#endif
{
    top_volume = new G4Tubs("top_volume",0.,pR,pDz,0.,360.);
    kTolerance=G4GeometryTolerance::GetInstance()->GetSurfaceTolerance();
    kHalfTolerance=kTolerance*.5;
    fR=min(fUpSurface->GetR(), fDownSurface->GetR());
    fZ1=fDownSurface->GetMinZ();
    fZ2=fUpSurface->GetMaxZ();
    fCylinder.SetR(fR);
    fCylinder.SetZ(fDownSurface->GetEdgeZ(), fUpSurface->GetEdgeZ());

    theLastPosition.set(DBL_MAX, DBL_MAX, DBL_MAX);
    theLastInside=kOutside;

    fSurfaceArea = 0.;

    __REG_ACTION("ShapeOfLens");
}

//__________________________________________________________________________________
ShapeOfLens::ShapeOfLens( __void__& a ) : G4VSolid(a),  fCubicVolume(0) {
    fSurfaceArea = 0.;
}

//__________________________________________________________________________________
ShapeOfLens::~ShapeOfLens() {
    // top_volume is deleted by Geant4
    //delete top_volume;
    if ( fUpSurface ) delete fUpSurface;
    if ( fDownSurface ) delete fDownSurface;
}

//__________________________________________________________________________________
EInside ShapeOfLens::Inside(const G4ThreeVector& p, int* surface, FSIntersectedSegment* intSeg) const {
    #ifdef DEBUG
    printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    printf("Inside called (%s)\n", GetName().data());
    printf("din pos(%f, %f, %f)\n",p.x(),p.y(),p.z());
    printf("rho=%f z=%f\n",p.perp(),p.z());
    #endif
    //if ( VectorIsZero(p) ) printf("zero\n");

    if ( CompareVectors(p, theLastPosition, kHalfTolerance) ) {
        if ( intSeg ) *intSeg=theLastInsideSegment;
        #ifdef DEBUG
        printf("vector is cached.\n");
        printf("return %s\n", theLastInside==kInside?"inside":(theLastInside==kOutside?"out":"surf"));
        #endif
        return theLastInside;
    }
    theLastInsideSegment.Reset();
    theLastPosition=p;

    double z = p.z();
    if ( z<fZ1-kHalfTolerance || z>fZ2+kHalfTolerance ) {
        #ifdef DEBUG
        printf("return out, for sure\n");
        #endif
        return theLastInside=kOutside;
    }

    double rho = p.perp();
    if ( rho > fR+kHalfTolerance ) {
        #ifdef DEBUG
        printf("return out, on side\n");
        #endif
        return theLastInside=kOutside;
    }


    EInside ins=fCylinder.Inside(rho, z, &theLastInsideSegment);
    if (ins==kSurface) {
        if (surface) *surface = 0;
        #ifdef DEBUG
        printf("return surf, cyl\n");
        #endif
        return theLastInside=kSurface;
    }

    EInside in = fDownSurface->Inside(rho, z, &theLastInsideSegment);
    if ( in==kInside ) {
        #ifdef DEBUG
        printf("return out, down\n");
        #endif
        return theLastInside=kOutside;
    }
    if ( in==kSurface) {
        if ( intSeg ) {
            *intSeg=theLastInsideSegment;
            if (intSeg->surf) printf("cyl err\n");
        }
        if (surface) *surface = 1;
        #ifdef DEBUG
        printf("return surface, down\n");
        #endif
        return theLastInside=kSurface;
    }

    in = fUpSurface->Inside(rho, z, &theLastInsideSegment);
    if ( intSeg && in==kSurface ) *intSeg=theLastInsideSegment;
    if (surface) *surface = 2;
    #ifdef DEBUG
    printf("return %s, up\n", in==kInside?"inside":(in==kOutside?"out":"in"));
    #endif
    return theLastInside=in;
}

//__________________________________________________________________________________
G4double ShapeOfLens::DistanceToIn(const G4ThreeVector& p, const G4ThreeVector& v) const {
    #ifdef DEBUG
    printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    printf("DistanceToIn(vec) called (%s)\n", GetName().data());
    printf("din pos(%.10g, %.10g, %.10g) dir(%.10g, %.10g, %.10g)\n",p.x(),p.y(),p.z(),v.x(),v.y(),v.z());
    printf("rho=%g z=%g, theta=%g phi=%g\n",p.perp(),p.z(), v.theta()/deg, v.phi()/deg);
    #endif

    double zp=p.z();
    double vz=v.z();

    theLastPositionOnSurface.set(kInfinity, kInfinity, kInfinity);
    theLastSurface=0;
    if ( (zp>=fZ2 && vz>=0.) || (zp<=fZ1 && vz<=0.) ) return kInfinity;

    double perp=p.perp();
    double smallSubst=0;
    FSIntersectedSegment intUp, intDown;
    theLastPositionOnSurface.set(kInfinity, kInfinity, kInfinity);
    if ( perp<(fR-kHalfTolerance) ) {
        if ( zp>=(fUpSurface->GetBottomZ(perp)-kHalfTolerance) ) {
            EInside in = fUpSurface->Inside(perp, zp);
            #ifdef DEBUG
            printf("in_up: %s\n", in==kSurface?"sur":( in==kOutside?"out":"in"));
            #endif
            if ( in!=kInside) {
                G4ThreeVector p1=p;
                if ( in==kSurface ) p1-=v*kHalfTolerance;
                double dd=fUpSurface->FindNearestIntersection(p1, v, &intUp);
                if ( in==kSurface ) dd-=kHalfTolerance;
                if( dd<=kHalfTolerance ) dd=smallSubst;;
                theLastPositionOnSurface=p+v*dd;
                theLastSurfaceSegment=intUp;
                theLastSurface=2;
                #ifdef DEBUG
                printf("dist up=%f\n",dd);
                if ( dd<kInfinity ) {
                    printf("next point (%f, %f, %f)\n",theLastPositionOnSurface.x(), theLastPositionOnSurface.y(), theLastPositionOnSurface.z());
                }
                #endif
                return dd;
            }
        }
        if ( zp<=(fDownSurface->GetUpperZ(perp)+kHalfTolerance) ) {
            EInside in = fDownSurface->Inside(perp, zp);
            #ifdef DEBUG
            printf("in_down: %s\n", in==kSurface?"sur":( in==kOutside?"out":"in"));
            #endif
            if ( in!=kOutside) {
                G4ThreeVector p1=p;
                if ( in==kSurface ) p1-=v*kHalfTolerance;
                double dd=fDownSurface->FindNearestIntersection(p, v, &intDown);
                if ( in==kSurface ) dd-=kHalfTolerance;
                if( dd<=kHalfTolerance ) dd=smallSubst;
                theLastPositionOnSurface=p+v*dd;
                theLastSurfaceSegment=intDown;
                theLastSurface=1;
                #ifdef DEBUG
                printf("dist down=%f\n",dd);
                if ( dd<kInfinity ) {
                    printf("next point (%f, %f, %f)\n",theLastPositionOnSurface.x(), theLastPositionOnSurface.y(), theLastPositionOnSurface.z());
                }
                #endif
                return dd;
            }
        }

        Warning_("Point is inside, but calling DistanceToIn.");
        int id=G4EventManager::GetEventManager()->
                               GetTrackingManager()->
                               GetTrack()->GetTrackID();
        printf("Event %i, %s\n", id, GetName().data());
        printf("pos(%.10g,%.10g,%.10g)\n", p.x(), p.y(), p.z());
        printf("dif(%.10g,%.10g,%.10g)\n", v.x(), v.y(), v.z());

        return kInfinity;
    }

    double cosalpha = p.x()*v.x()+p.y()*v.y();
    if ( cosalpha>=.0 ) return kInfinity;
    // calculate distance to cylinder
    FSIntersectedSegment intCyl;
    double dist_cyl  = fCylinder.FindNearestIntersection(p, v, &intCyl);
    double dist_up   = fUpSurface->FindNearestIntersection(p, v, &intUp);
    double dist_down = fDownSurface->FindNearestIntersection(p, v, &intDown);

    #ifdef DEBUG
    printf("dc: %e \tdu: %e \t dd:%e\n",dist_cyl,dist_up,dist_down);
    #endif

    double mindist;
    if ( dist_up<dist_down ) {
        if ( dist_up<dist_cyl ){
            mindist=dist_up;
            theLastSurfaceSegment=intUp;
            theLastSurface=2;
        }
        else {
            mindist=dist_cyl;
            theLastSurfaceSegment=intCyl;
            theLastSurface=0;
        }
    }
    else {
        if ( dist_down<dist_cyl ){
            mindist=dist_down;
            theLastSurfaceSegment=intDown;
            theLastSurface=1;
        }
        else {
            mindist=dist_cyl;
            theLastSurfaceSegment=intCyl;
            theLastSurface=0;
        }
    }
    mindist=mindist>kTolerance?dist_cyl:smallSubst;
    theLastPositionOnSurface=p+v*mindist;
    return mindist;
}

//__________________________________________________________________________________
G4double ShapeOfLens::DistanceToIn(const G4ThreeVector& p) const{
    #ifdef DEBUG
    printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    printf("DistanceToIn(pos) called (%s)\n", GetName().data());
    printf("din pos(%f, %f, %f)\n",p.x(),p.y(),p.z());
    printf("rho=%f z=%f\n",p.perp(),p.z());
    #endif
    double zp=p.z();
    if ( zp<=fCylinder.GetMinZ() ) return fDownSurface->DistanceToSurfSafeDown(p);
    if ( zp>=fCylinder.GetMaxZ() ) return fUpSurface->DistanceToSurfSafeUp(p);
    return fabs(p.perp()-fR);
}

//__________________________________________________________________________________
G4double ShapeOfLens::DistanceToOut(const G4ThreeVector& p,
                           const G4ThreeVector& v,
                           const G4bool calcNorm,
                                 G4bool *validNorm,
                                 G4ThreeVector *n) const {
    #ifdef DEBUG
    printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    printf("DistanceToOut(vec) called (%s)\n", GetName().data());
    printf("dout (%f, %f, %f) (%f, %f, %f)\n",p.x(),p.y(),p.z(),v.x(),v.y(),v.z());    printf("rho=%f z=%f\n",p.perp(),p.z());
    #endif
    if ( calcNorm ) *validNorm = true;

    G4ThreeVector p1=p;
    double dist_shift(0.);
    double perp=p.perp();
    EInside in_up = fUpSurface->Inside(perp,p.z());
    EInside in_down = fDownSurface->Inside(perp,p.z());
    #ifdef DEBUG
    printf("in_up: %s\n",in_up==kSurface?"sur":(in_up==kOutside?"out":"in"));
    printf("in_down: %s\n",in_down==kSurface?"sur":(in_down==kOutside?"out":"in"));
    #endif
    if ( in_up==kSurface || in_down==kSurface ) {
        #ifdef DEBUG
        printf("point is on surface, shift it by kHalfTolerance\n");
        #endif
        p1+=kHalfTolerance*v;
        dist_shift+=kHalfTolerance;
    }

    #ifdef DEBUG
    printf("asking up\n");
    #endif
    FSIntersectedSegment intsegUp, intsegDown;
    double dist_up = fUpSurface->FindNearestIntersection(p1, v, &intsegUp);
    #ifdef DEBUG
    printf("asking down\n");
    #endif
    double dist_down = fDownSurface->FindNearestIntersection(p1, v, &intsegDown);

    #ifdef DEBUG
      printf("du=%e dd=%e\n",dist_up,dist_down);
      double du=dist_up+dist_shift;
      double dd=dist_down+dist_shift;
      printf("du(s)=%e dd=%e\n",du,dd);
      if ( du<kInfinity ) {
          G4ThreeVector newp=p+v*du;
          printf("next point (up) (%f, %f, %f)\n",newp.x(), newp.y(), newp.z());
      }
      if ( dd<kInfinity ) {
          G4ThreeVector newp=p+v*dd;
          printf("next point (down) (%f, %f, %f)\n",newp.x(), newp.y(), newp.z());
      }
    #endif

    if ( dist_up<kInfinity || dist_down<kInfinity ) {
        if ( dist_up<dist_down ) {
            if ( calcNorm ) {
                intsegUp.Normal(p1+v*dist_up, +1., n);
                #ifdef SLANTWALLS
                if ( isSlant && n->z()==0. ) CorrectCylinderNormal(n);
                #endif
                #ifdef DEBUG
                printf("normal is (%f, %f, %f)\n",n->x(), n->y(), n->z());
                #endif
            }
            dist_shift+=dist_up;
            return dist_shift<kHalfTolerance?kHalfTolerance:dist_shift+kHalfTolerance*0.5;
        }

        if ( dist_down<kInfinity ) {
            if ( calcNorm ) {
                intsegDown.Normal(p1+v*dist_down, -1., n);
                #ifdef SLANTWALLS
                if ( isSlant && n->z()==0.) CorrectCylinderNormal(n);
                #endif
                #ifdef DEBUG
                printf("normal is (%f, %f, %f)\n",n->x(), n->y(), n->z());
                #endif
            }
            dist_shift+=dist_down;
            return dist_shift<kHalfTolerance?kHalfTolerance:dist_shift+kHalfTolerance*0.5;
        }
    }

    double dist;
    G4ThreeVector pp;
    if ( calcNorm ) {
        FSIntersectedSegment intCyl;
        dist=fCylinder.FindNearestIntersection(p1, v, &intCyl);
        pp=p1+v*dist;
        intCyl.Normal(pp, -1., n);
        #ifdef SLANTWALLS
        if ( isSlant ) CorrectCylinderNormal(n);
        #endif
    }
    else {
        dist=fCylinder.FindNearestIntersection(p1, v);
        pp=p1+v*dist;
    }

    if ( dist==kInfinity ) {
        if ( dist_shift ) return dist_shift<kHalfTolerance?kHalfTolerance:dist_shift;
        else {
            #ifdef DEBUG_WNF
                printf("shift is %g\n", dist_shift);
                if ( fUpSurface->getDebug() ) exit(1);
                fUpSurface->setDebug();
                fDownSurface->setDebug();
                DistanceToOut(p, v, calcNorm, validNorm, n);
            #endif
            Warning_("Way out not found.");
            int id=G4EventManager::GetEventManager()->
                                   GetTrackingManager()->
                                   GetTrack()->GetTrackID();
            printf("Event %i, %s\n", id, GetName().data());
            printf("pos(%.10g,%.10g,%.10g)\n", p.x(), p.y(), p.z());
            printf("dif(%.10g,%.10g,%.10g)\n", v.x(), v.y(), v.z());
        }
    }
    dist+=dist_shift;
    return dist<kHalfTolerance?kHalfTolerance:dist+kHalfTolerance*0.5;
}

//__________________________________________________________________________________
G4double ShapeOfLens::DistanceToOut(const G4ThreeVector& p) const{
    #ifdef DEBUG
    printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    printf("DistanceToOut(pos) called (%s)\n", GetName().data());
    printf("din pos(%f, %f, %f)\n",p.x(),p.y(),p.z());
    printf("rho=%f z=%f\n",p.perp(),p.z());
    #endif
    double d1=fDownSurface->DistanceToSurfSafeUp(p);
    if ( d1<kTolerance ){
        #ifdef DEBUG
            printf("return dist to up: %g\n",d1);
        #endif
        return d1;
    }
    double d2=fUpSurface->DistanceToSurfSafeDown(p);
    if ( d2<kTolerance ){
        #ifdef DEBUG
            printf("dist to up: %g\n", d1);
            printf("return dist to down: %g\n",d2);
        #endif
        return d2;
    }
    double d3=fabs(fR-p.perp());

    #ifdef DEBUG
        printf("dist to up: %g\n", d1);
        printf("dist to down: %g\n", d2);
        printf("dist to cyl: %g\n", d3);
        printf("return: %g\n", min(min(d1,d2),d3));
    #endif
    return min(min(d1,d2),d3);
}

//#define DEBUG_SAVENORMAL
//__________________________________________________________________________________
G4ThreeVector ShapeOfLens::SurfaceNormal( const G4ThreeVector& p ) const {
    #ifdef DEBUG
    printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    printf("SurfaceNormal called (%s)\n", GetName().data());
    printf("p (%f, %f)\n",p.perp(),p.z());
    #endif
    int surface(theLastSurface);
    FSIntersectedSegment intseg;
    G4ThreeVector norm(1.,0.,0.);
    if ( CompareVectors(p, theLastPositionOnSurface, kHalfTolerance) ){
        intseg=theLastSurfaceSegment;
        #ifdef DEBUG
        printf("Point is saved, using the last segment normal.\n");
        #endif
    }
    else {
        EInside ins=this->Inside(p, &surface, &intseg);
        if ( ins!=kSurface ){
            Warning_("Calculating SurfaceNormal not on surface.");
            int id=G4EventManager::GetEventManager()->
                                   GetTrackingManager()->
                                   GetTrack()->GetTrackID();
            printf("Event %i, %s\n", id, GetName().data());
            double rho=p.perp();
            printf("pos (%.10g,%.10g,%.10g), rho=%.10g\n", p.x(), p.y(), p.z(), rho);
            double z1=fUpSurface->PointOnSurf(rho), z2=fDownSurface->PointOnSurf(rho);
            printf("dz1 %g, dz2 %g: z1, z2 (%g, %g)\n", p.z()-z1,p.z()-z2, z1, z2);
            return norm;
        }

        #ifdef DEBUG
        printf("Point is not saved. Dist is %g\n", (p-theLastPositionOnSurface).mag() );
        #endif
    }
    #ifdef DEBUG
    printf("    Point: %g %g %g\n",p.x(), p.y(), p.z());
    printf("    Saved point: %g %g %g\n",theLastPositionOnSurface.x(), theLastPositionOnSurface.y(), theLastPositionOnSurface.z());
    printf("    Saved segment: type %2i, r(%g, %g), z(%g, %g)\n", intseg.type, intseg.r1, intseg.r2, intseg.z1, intseg.z2);
    #endif
    intseg.Normal(p, surface<2?-1.:1., &norm);
    #ifdef DEBUG
    double  mag=norm.mag();
    printf("Returned normal: (%g, %g, %g) (%g, %g) len=%g\n", norm.x(), norm.y(), norm.z(), norm.theta()/deg, norm.phi()/deg, mag);
    if ( fabs(mag-1.)>kHalfTolerance ) printf("ERROR: not unitary.\n");
    #endif

    #ifdef SLANTWALLS
    if ( isSlant && norm.z()==0. ) CorrectCylinderNormal(&norm);
    #endif

    return norm;
}

//__________________________________________________________________________________
#ifdef SLANTWALLS
void ShapeOfLens::CorrectCylinderNormal(G4ThreeVector* n) const {
    n->setTheta( CLHEP::RandGauss::shoot(90.*deg, theSlantSigma) );
    #ifdef DEBUG_SLANTWALLS
        printf("%f\n",n->getTheta()/deg);
    #endif
}
#endif

//__________________________________________________________________________________
G4double ShapeOfLens::GetCubicVolume(){
  if(fCubicVolume == 0.)fCubicVolume=top_volume->GetCubicVolume();
  return fCubicVolume;
}

//__________________________________________________________________________________
G4double ShapeOfLens::GetSurfaceArea() {
  if(fSurfaceArea == 0.)fSurfaceArea=top_volume->GetSurfaceArea();
  return fSurfaceArea;
}

//__________________________________________________________________________________
void ShapeOfLens::DescribeYourselfTo(G4VGraphicsScene& scene) const{
    top_volume->DescribeYourselfTo(scene);
}
    
//__________________________________________________________________________________
G4Polyhedron* ShapeOfLens::CreatePolyhedron() const {
  return top_volume->CreatePolyhedron();
}

//__________________________________________________________________________________
G4Polyhedron* ShapeOfLens::GetPolyhedron () const {
  return top_volume->GetPolyhedron();
}

//__________________________________________________________________________________
G4NURBS*      ShapeOfLens::CreateNURBS() const {
  return top_volume->CreateNURBS();
}

//__________________________________________________________________________________
G4ThreeVector ShapeOfLens::GetPointOnSurface() const{
    return top_volume->GetPointOnSurface();
}

//__________________________________________________________________________________
G4bool ShapeOfLens::CalculateExtent(const EAxis pAxis,
                           const G4VoxelLimits& pVoxelLimit,
                           const G4AffineTransform& pTransform,
                                 G4double& pmin, G4double& pmax) const {
    return top_volume->CalculateExtent(pAxis,pVoxelLimit,pTransform,pmin,pmax);
}
