#include "G4Detector.hh"

#include "Detector.hh"
#include "EusoMapping.hh"
#include "DetectorGeometry.hh"
#include "VirtualDetectorTransportManager.hh"
#include "ElectronicsFactory.hh"
#include "ERunParameters.hh"
#include "ERunParsPixelMapFiller.hh"
#include "ERunParsPmtsFiller.hh"
#include "EusoElectronics.hh"
#include "OpticsFactory.hh"
#include "Pupil.hh"
#include "ParentPhoton.hh"
#include "Photon.hh"

#include <TMath.h>
#include <TFile.h>
using namespace sou;

//______________________________________________________________________________
G4Detector::G4Detector() : fElectronics(0) {
    //
    // Constructor
    //
    // FIXME: call the factories to get the right associate child and parent
    // classes in config
    ElectronicsFactory::Get();
    OpticsFactory::Get();
    EusoMapping::Get();
    Build();
    SetSender("G4Detector");
    MsgForm(EsafMsg::Info,"Creating G4Detector.");
}

//______________________________________________________________________________
G4Detector::~G4Detector() {
    //
    // Destructor
    //
    if (fGeometry)
        delete fGeometry;
    if (fDetectorTransporter)
        delete fDetectorTransporter;
    ElectronicsFactory::Get()->DeleteEusoElectronics();
}

//______________________________________________________________________________
void G4Detector::Build() {
    //
    // Build or re-build the detector
    //
    fGeometry = new DetectorGeometry();
    fGeometry->SetFoV(Pi()/6.);

    Msg(EsafMsg::Info)  << "Creating G4DetectorTransporter." << MsgDispatch;
    fDetectorTransporter = OpticsFactory::Get()->MakeTransportManager();
    fElectronics = ElectronicsFactory::Get()->Build();

    Double_t scale = Conf()->GetNum("G4Detector.fScaleFactor");
    Double_t radius;
    if ( Conf()->GetBool("G4Detector.fFixMaxRadius") )
         radius = Conf()->GetNum("G4Detector.fMaxRadius");
    else radius = fDetectorTransporter->GetOuterRadius();

    fGeometry->SetRadius(scale*radius);
    fDetectorTransporter->SetScaleFactor(scale);

    // save run parameters
    if ( EEvent::GetCurrent() && EEvent::GetCurrent()->GetRunPars() ) {
        ERunParameters *runpars = EEvent::GetCurrent()->GetRunPars();

        fElectronics->SetDetectorScaleFactor(scale);
        fElectronics->BuildBackgroundChipDist();

        if ( runpars ) {
            // get from EusoMapping the config
            ERunParsPixelMapFiller pxfiller;
            runpars->Fill(pxfiller);

            ERunParsPmtsFiller pmtfiller( fElectronics );
            runpars->Fill(pmtfiller);
        }
    }
    fDetectorTransporter->GetFocalPlane()->SetElectronics(fElectronics);
}

void G4Detector::Reset() {
    //
    // Reset electronics
    //

    if ( fElectronics )
        fElectronics->Reset();
}

Telemetry* G4Detector::Get( PhotonsOnPupil* photons ) {
    //
    // Simulate one event and return a pointer to a Telemetry object
    //


    // wrapper to track the photons onto the real pupil
    
//     TFile f("output/event.root","RECREATE");
    
    Pupil pupil( photons, GetGeometry());
/*		
		TTree* t = new TTree("t","tt");
		Photon* photon =0;
		ParentPhoton* pp =0;*/

// 		TVector3 vpos(0.,0.,0.);
// 		TVector3 vdir(0.,0.,0.);
// 		
// 		t->Branch("position","TVector3",&vpos);
// 		t->Branch("direction","TVector3",&vdir);
// 		t->Branch("wl",&wl,"w/D");
// 		t->Branch("flag",&fl,"flag/I");

// 		t->Branch("photon","ParentPhoton",&pp);
    if(pupil.GetNphotons()==0) Msg(EsafMsg::Warning)  << "NO PHOTONS ON PUPIL !!!"<< MsgDispatch;
    else {Msg(EsafMsg::Info)  << "PHOTONS ON PUPIL: "<<pupil.GetNphotons()<< MsgDispatch;
// 			photon = pupil.GetPhoton();
// 			pp=photon->parent;
// 			printf("www  %g\n",pp->W());
// 			int id=0;
// 			
// 		  while(id<pupil.GetNphotons()){
// 				photon->id = id;
// 				t->Fill();
// 				
// 				photon = pupil.GetPhoton();
// 				if(photon){
// 					pp=photon->parent;
// 					printf("www  %g\n",pp->W());
// 					
// 				}
// 				id++;
// 	//			photon->Dump();
// 			
// 			}
// 			f.Write();
// 			f.Close();
		}

    // transport photons from pupil to pmts
    fDetectorTransporter->Go( &pupil );

    Double_t tBegin = fDetectorTransporter->GetTimeFirstPhoton();
    Double_t tEnd = fDetectorTransporter->GetTimeLastPhoton();

    MsgForm(EsafMsg::Info,"Photon Time interval (ns):\t\tSTART=%ld  END=%ld",
            (long int)(tBegin), (long int)tEnd);

    // simulate electronics and trigger
    if ( !fElectronics->Simulate( tBegin, tEnd ) )
        return (Telemetry*)0;

    // return telemetry object
    return fElectronics->Data();

}


//______________________________________________________________________________
Bool_t G4Detector::ClearMemory() {
    //
    // Physically release the memory allocated by the arrays of this object
    //

    fElectronics->ClearMemory();

    return kTRUE;
}
