#include "G4EusoGeometryShape.hh"

#include "FresnelSurface.h"
#include "FresnelSurfaceHelper.h"
#include "ParabolicSurface.h"
#include "SphericSurface.h"
#include "FlatSurface.h"
#include "ShapeOfLens.h"
#include "MaterialParameters.hh"

#include <G4AssemblyVolume.hh>
#include <G4Box.hh>
#include <G4IntersectionSolid.hh>
#include <G4Cons.hh>
#include <G4DisplacedSolid.hh>
#include <G4Element.hh>
#include <G4ElementTable.hh>
#include <G4GeometryTolerance.hh>
#include <G4IntersectionSolid.hh>
#include <G4LogicalBorderSurface.hh>
#include <G4LogicalSkinSurface.hh>
#include <G4LogicalVolume.hh>
#include <G4Material.hh>
#include <G4MaterialTable.hh>
#include <G4OpBoundaryProcess.hh>
#include <G4PVPlacement.hh>
#include <G4Paraboloid.hh>
#include <G4Sphere.hh>
#include <G4SubtractionSolid.hh>
#include <G4Torus.hh>
#include <G4Transform3D.hh>
#include <G4Tubs.hh>

#include <TEnv.h>
#include <TGraph.h>
#include <TGeoManager.h>
#include <TSystem.h>
#include <TF1.h>
#include <TFile.h>
#include <cmath>
#include <iostream>
using namespace std;
using namespace TMath;
using namespace G4FresnelLens;

#include "Ntrace_optics.hh"
using namespace NTraceLens;
//using std::sqrt;
//using std::cout;
//using std::flush;

// #include <Geant4GM/volumes/Factory.h>
// #include <RootGM/volumes/Factory.h>

#include <stdlib.h>

// #define DISABLE_1_LENSphys
// #define DISABLE_2_LENSphys
// #define DISABLE_3_LENSphys

G4EusoGeometryShape::G4EusoGeometryShape()
{
    fGeomTolerance=G4GeometryTolerance::GetInstance()->GetSurfaceTolerance();
    expHall_x = expHall_y = 2000.*mm;
    expHall_z = 4000.*mm;
    fit = Conf()->GetBool("G4EusoGeometryShape.UseFitTransmittivity");
    CYTOP=0;
    PMMA=0;
    fOpticalSurfaceC=0;
    fOpticalSurfaceP=0;


}

G4EusoGeometryShape::~G4EusoGeometryShape(){
    if(CYTOP)delete CYTOP;
    if(PMMA)delete PMMA;
    if(fOpticalSurfaceC)delete fOpticalSurfaceC;
    if(fOpticalSurfaceP)delete fOpticalSurfaceP;

}


G4VPhysicalVolume* G4EusoGeometryShape::Construct(){
    lens_dir =fOpticalSystemParam->lens_dir;
#ifdef DEBUG
    printf("%p %p\n",fOpticalSystemParam, fOpticalSystemParam->lens_dir);
    printf("%s\n",lens_dir);
    printf("%s \n", fOpticalSystemParam->name);
#endif
    G4double  a, z, density;
    G4int nelements;
    /// elements
    G4Element* H = new G4Element("Hydrogen", "H", z=1 , a=1.00794*g/mole);
    G4Element* O = new G4Element("Oxygen"  , "O", z=8 , a=16.00*g/mole);
    G4Element* C = new G4Element("Carbon", "C", z=6 , a=12.0107*g/mole);
    G4Element* P = new G4Element("Phosphorus", "P", z=15 , a=30.973761*g/mole);

    G4Material* Tributylphosphine = new G4Material("Tributylphosphine", density=0.81*g/cm3, nelements=3);
    Tributylphosphine->AddElement(P, 1);
    Tributylphosphine->AddElement(C, 12);
    Tributylphosphine->AddElement(H, 27);

    G4Material* Dibutylphosphine = new G4Material("Dibutylphosphine", density=0.81*g/cm3, nelements=3);
    Dibutylphosphine->AddElement(P, 1);
    Dibutylphosphine->AddElement(C, 8);
    Dibutylphosphine->AddElement(H, 19);

    G4Material* TributylphosphineOxide = new G4Material("TributylphosphineOxide", density=0.81*g/cm3, nelements=4);
    TributylphosphineOxide->AddElement(P, 1);
    TributylphosphineOxide->AddElement(C, 12);
    TributylphosphineOxide->AddElement(H, 27);
    TributylphosphineOxide->AddElement(O, 1);
    ///material
    CYTOP = new G4Material("CYTOP", density=2.03*g/cm3, nelements=3);
    CYTOP->AddMaterial(Tributylphosphine,0.98);
    CYTOP->AddMaterial(TributylphosphineOxide,0.005);
    CYTOP->AddMaterial(Dibutylphosphine,0.015);

    PMMA = new G4Material("PMMA", density=1.19*g/cm3, nelements=3);
    PMMA->AddElement(O,2);
    PMMA->AddElement(H,8);
    PMMA->AddElement(C,5);
    // Vacuum-world; Vacuum1-FS; Vacuum2-diffraction
    G4Material* Vacuum =
    new G4Material("Galactic", z=1., a=1.01*g/mole,density= universe_mean_density,kStateGas, 2.73*kelvin, 3.e-18*pascal);
    G4Material* Vacuum1 =
    new G4Material("Galactic1", z=1., a=1.01*g/mole,density= 10.*g/cm3);
    G4Material* Vacuum2 =
    new G4Material("Galactic", z=1., a=1.01*g/mole,density= universe_mean_density,kStateGas, 2.73*kelvin, 3.e-18*pascal);

    //FIXME
    //Iris material, now IRIS is just black body, cover also is a black body
    G4Material* fe_mat= new G4Material("fe_mat", z=26. , a=55.845*g/mole, density=7.87*g/cm3);

    ///Material Properties Table

    const G4int nEntries = 18;
    const double lambdas1[nEntries]={245.,275.,305., 310.,320., 330., 335., 340.,
                                    350., 360., 370., 380., 390., 400., 410., 420., 435., 546.} ;

    double cnst=hbarc*twopi/eV;
    G4double PhotonEnergy1[nEntries];
    for (int i(0); i<nEntries; i++){
        PhotonEnergy1[i]=cnst/lambdas1[nEntries-i-1];
    }

    G4double RefractiveIndex1[nEntries] =
            { 1.00, 1.00, 1.00, 1.00, 1.00, 1.00, 1.00,
              1.00, 1.00, 1.00, 1.00, 1.00, 1.00, 1.00,
              1.00, 1.00, 1.00, 1.00};

     G4double  EUSO_DINDEX[nEntries] =
            { 0., 0., 0., 0.,0., 0., 0.,
              0., 0., 0., 0.,0., 0., 0.,
              0., 0., 0., 0.};

    G4MaterialPropertiesTable* myMPT1 = new G4MaterialPropertiesTable();
    myMPT1->AddProperty("RINDEX", PhotonEnergy1, RefractiveIndex1, nEntries);
    Vacuum->SetMaterialPropertiesTable(myMPT1);
    Vacuum1->SetMaterialPropertiesTable(myMPT1);

    fOpticalSurfaceC = new G4OpticalSurface("vacuum_cytop_os");
    fOpticalSurfaceC->SetType(dielectric_dielectric);
    fOpticalSurfaceC->SetModel(unified);
    fOpticalSurfaceC->SetFinish(polished);

    fOpticalSurfaceP = new G4OpticalSurface("vacuum_pmma_os");
    fOpticalSurfaceP->SetType(dielectric_dielectric);
    fOpticalSurfaceP->SetModel(unified);
    //fOpticalSurfaceP->SetFinish(polishedbackpainted);
    fOpticalSurfaceP->SetFinish(polished);


    MaterialParameters* mp1 = new MaterialParameters();
    MaterialParameters* mp2 = new MaterialParameters();
    char m1mat[50];
    char m2mat[50];

    SetMaterial(2,mp1,m1mat);  //2 and 3 number of materials, 1&2-air,vacuum. 2&3-materials of lenses
    SetMaterial(3,mp2,m2mat);


    /// material properties to use with OpESAFDiffraction
    /// there will be thin vacuum layer after 2-d lens with normal vacuum properties
    /// but special properties of PMMA
    /// so G4OpBoundaryProcess will see vacuum and do nothing
    /// and OpESAFDiffraction will see PMMA and do special diffraction

    G4MaterialPropertiesTable* myMPT3a = new G4MaterialPropertiesTable();
    myMPT3a->AddProperty("RINDEX", PhotonEnergy1, RefractiveIndex1, nEntries);
    myMPT3a->AddProperty("EUSO_DINDEX",PhotonEnergy1,EUSO_DINDEX,nEntries);
    //diffractive vacuum
    //for the EUSO diffraction it has RINDEX properties as second lens
    char* matdoe = fOpticalSystemParam->material[fOpticalSystemParam->surface[5].matid].name;
    G4MaterialPropertiesTable* mprop=0;

    if(!strcmp(matdoe,m1mat)){
        mprop=mp1->GetTable();
     }
    else if(!strcmp(matdoe,m2mat)){
        mprop=mp2->GetTable();
    }
    else cerr<<"Unknown doe material"<<endl;

    G4MaterialPropertyVector *mv = mprop->GetProperty ("RINDEX");
    myMPT3a->AddProperty ("EUSO_RINDEX",mv);
    Vacuum2->SetMaterialPropertiesTable(myMPT3a);

    //volumes
    ///world
    G4Box* expHall = new G4Box("World",expHall_x,expHall_y,expHall_z);
    G4LogicalVolume* expHall_log = new G4LogicalVolume(expHall,Vacuum,"World",0,0,0);
    G4VPhysicalVolume* expHall_phys = new G4PVPlacement(0,G4ThreeVector(),expHall_log,"World",0,false,0);

    //intersection volume for lenses
    double r_wall=fOpticalSystemParam->r_wall;
    double cut =fOpticalSystemParam->surface[1].r_cut;
    G4Box* box = new G4Box("box",(r_wall+10.)*mm,cut*mm,1.*m);
    
   
    /// 1-st lens

    double c_rad, c_z1, c_z2, c_dz;
    char* name=fOpticalSystemParam->name;
    

    double surf1=fOpticalSystemParam->surface[1].Sz;
    double surf2=fOpticalSystemParam->surface[2].Sz;
    double thin= surf2 - surf1;
    double Rc1=fOpticalSystemParam->surface[1].Rc;
    double Rc2=fOpticalSystemParam->surface[2].Rc;
    char* mat = fOpticalSystemParam->material[fOpticalSystemParam->surface[2].matid].name;
 //
    //const char* lens_file="output/surfaces.root";
    FresnelSurfaceHelper* shelper=new FresnelSurfaceHelper();
    double filter=0.5;
    if ( !strcmp(name,"CPC_2007") ){
        filter=DBL_MAX;
        //printf("Disabling filter for (%s)\n", name);
    }

//     const char* lens_file="output/surfaces.root";
#ifndef DISABLE_1_LENS
    ParametricSurface* uplimit=new SphericSurface(Rc1*mm,0.,r_wall);
    ParametricSurface* downlimit=new SphericSurface(Rc1*mm,0.,r_wall);

    shelper->SetOptions(0.*mm, -1., -filter);
    shelper->Read(Form("%s/S1_%s.dat",lens_dir,name));
    FresnelSurface* downsurf1=shelper->CreateFresnelLens(downlimit, uplimit);

    uplimit=new SphericSurface(Rc2*mm,0.,r_wall);
    downlimit=new SphericSurface(Rc2*mm,0.,r_wall);

    shelper->SetOptions(thin*mm, -1., filter);
    shelper->Read(Form("%s/S2_%s.dat",lens_dir,name));
    FresnelSurface* upsurf1=shelper->CreateFresnelLens(downlimit, uplimit);
    ShapeOfLens* lens1_solid = new ShapeOfLens("FresnelLens_1", (thin*0.5+0.5)*mm, (r_wall+1.)*mm, upsurf1, downsurf1);

    G4IntersectionSolid* intersec1 = new G4IntersectionSolid("box*FresnelLens_1", box, lens1_solid);
    G4LogicalVolume* lens1_log = new G4LogicalVolume(intersec1,!strcmp(mat,"CYTOP")?CYTOP:PMMA, "FresnelLens_1",0,0,0);
    
    #ifndef DISABLE_1_LENSphys
    G4VPhysicalVolume* lens1_phys = new G4PVPlacement(0, G4ThreeVector(0.,0.,surf1), lens1_log, "FresnelLens_1", expHall_log,false,0);
    /*G4LogicalBorderSurface diffr_surf =*/
    new G4LogicalBorderSurface("diffractor_surf1",lens1_phys,expHall_phys,!strcmp(mat,"CYTOP")?fOpticalSurfaceC:fOpticalSurfaceP);
    #endif
#endif

    ///2-d lens
#ifndef DISABLE_2_LENS

    surf1 = fOpticalSystemParam->surface[4].Sz;
    surf2 = fOpticalSystemParam->surface[5].Sz;
    thin = surf2 -surf1;
    Rc1 = fOpticalSystemParam->surface[4].Rc;
    Rc2 = fOpticalSystemParam->surface[5].Rc;
    mat = fOpticalSystemParam->material[fOpticalSystemParam->surface[5].matid].name;
    cut =fOpticalSystemParam->surface[4].r_cut;

    if(Rc2>1.0e10){
         uplimit=new FlatSurface();
         downlimit=new FlatSurface();
    }
    else{
        uplimit=new SphericSurface(Rc1*mm,0.,r_wall);
        downlimit=new SphericSurface(Rc1*mm,0.,r_wall);
    }
    shelper->SetOptions(0., -1., -filter);
    shelper->Read(Form("%s/S4_%s.dat",lens_dir,name));
    FresnelSurface* downsurf2=shelper->CreateFresnelLens(downlimit, uplimit);
    double rsurf5=downsurf2->GetR();

    ParametricSurface* upsurf2;

    if(Rc2>1.0e10){
         upsurf2 = new FlatSurface((thin)*mm,rsurf5*mm);
    }
    else{
        upsurf2=new SphericSurface(Rc2*mm,(thin+Rc2)*mm,rsurf5*mm);
    }
    ShapeOfLens* lens2_solid = new ShapeOfLens("FresnelLens_2", (thin*0.5+0.5)*mm, (rsurf5+1.)*mm, upsurf2, downsurf2);
    G4IntersectionSolid* intersec2 = new G4IntersectionSolid("box*FresnelLens_2", box, lens2_solid);
    
    G4LogicalVolume* lens2_log = new G4LogicalVolume(intersec2,!strcmp(mat,"CYTOP")?CYTOP:PMMA, "FresnelLens_2",0,0,0);

    double diffr_rad=rsurf5; //FIXME: diffractive surface is shorter than surf5
    double diffr_dz=1.e-8;
    G4VSolid* euso_diffractor;

    if(Rc2>1.0e10){
         euso_diffractor=new G4Tubs("euso_diffractor_tubs", 0., diffr_rad, diffr_dz,0.*deg,360.*deg);
    }
    else {
        ParametricSurface* upsurf_diffr = new SphericSurface(Rc2*mm,Rc2*mm+diffr_dz,rsurf5*mm);
        ParametricSurface* downsurf_diffr = new SphericSurface(Rc2*mm,Rc2*mm-diffr_dz,rsurf5*mm);
        euso_diffractor=new ShapeOfLens("euso_diffractor_par", diffr_dz, diffr_rad, upsurf_diffr, downsurf_diffr);
	
    }
    G4IntersectionSolid* intersec_diffr = new G4IntersectionSolid("box*euso_diffractor_par", box, euso_diffractor);
    G4LogicalVolume* euso_diffractor_log = new G4LogicalVolume(intersec_diffr,Vacuum2,"euso_diffractor",0,0,0);

    #ifndef DISABLE_2_LENSphys
    G4VPhysicalVolume* lens2_phys = new G4PVPlacement(0, G4ThreeVector(0.,0.,surf1), lens2_log, "FresnelLens_2", expHall_log,false,0);
    //G4VPhysicalVolume* euso_diffractor_phys =
            new G4PVPlacement(0,G4ThreeVector(0,0,thin+diffr_dz+fGeomTolerance+surf1),
                                              euso_diffractor_log,"euso_diffractor",expHall_log,false,0);

    /*G4LogicalBorderSurface diffr_surf =*/
    new G4LogicalBorderSurface("diffractor_surf21",lens2_phys,expHall_phys,!strcmp(mat,"CYTOP")?fOpticalSurfaceC:fOpticalSurfaceP);
    #endif
#endif

    /// 3-d Lens
#ifndef DISABLE_3_LENS
    surf1 = fOpticalSystemParam->surface[6].Sz;
    surf2 = fOpticalSystemParam->surface[7].Sz;
    thin = surf2 -surf1;
    Rc1 = fOpticalSystemParam->surface[6].Rc;
    Rc2 = fOpticalSystemParam->surface[7].Rc;
    mat = fOpticalSystemParam->material[fOpticalSystemParam->surface[7].matid].name;
    cut =fOpticalSystemParam->surface[6].r_cut;

     uplimit=new SphericSurface(Rc1,0.,r_wall);
     downlimit=new SphericSurface(Rc1,0.,r_wall);
     shelper->SetOptions(0., -1., -filter);
     shelper->Read(Form("%s/S6_%s.dat",lens_dir,name));
     FresnelSurface* downsurf3=shelper->CreateFresnelLens(downlimit, uplimit);

     uplimit=new SphericSurface(Rc2,0.,r_wall);
     downlimit=new SphericSurface(Rc2,0.,r_wall);
     shelper->SetOptions(thin*mm, -1., -filter);
     shelper->Read(Form("%s/S7_%s.dat",lens_dir,name));
     FresnelSurface* upsurf3=shelper->CreateFresnelLens(downlimit, uplimit);

     ShapeOfLens* lens3_solid = new ShapeOfLens("FresnelLens_3",(thin*0.5+0.5)*mm, (r_wall+1.)*mm, upsurf3, downsurf3);
     G4IntersectionSolid* intersec3 = new G4IntersectionSolid("box*FresnelLens_3", box, lens3_solid);

     G4LogicalVolume* lens3_log = new G4LogicalVolume(intersec3,!strcmp(mat,"CYTOP")?CYTOP:PMMA, "FresnelLens_3",0,0,0);

     lens3_solid->GetCylinderLimits(c_rad, c_z1, c_z2);
     c_dz=(fabs(c_z2)+fabs(c_z1));


    #ifndef DISABLE_3_LENSphys
    G4VPhysicalVolume* lens3_phys = new G4PVPlacement(0, G4ThreeVector(0.,0.,surf1), lens3_log, "FresnelLens_3", expHall_log,false,0);
    /*G4LogicalBorderSurface diffr_surf =*/
    new G4LogicalBorderSurface("diffractor_surf3",lens3_phys,expHall_phys,!strcmp(mat,"CYTOP")?fOpticalSurfaceC:fOpticalSurfaceP);
    #endif
#endif
    delete shelper;

    /// fs
    double fs_dz=10.*mm;
    double fs_pos= ((fOpticalSystemParam->FS_OFFSET-surf2)*0.5+surf2)*mm;//surf1+c_dz+15.*mm+fs_dz;
    double fs_rad=fOpticalSystemParam->r_fs*mm;

    G4Tubs* absorber_cyl=new G4Tubs("absorber",0.,fs_rad,fs_dz,0.*deg,360.*deg);
    G4IntersectionSolid* intersec_fs = new G4IntersectionSolid("box*absorber", box, absorber_cyl);
    G4LogicalVolume* absorber_cyl_log = new G4LogicalVolume(intersec_fs,Vacuum1,"absorber",0,0,0);
    //G4VPhysicalVolume* absorber_cyl_phys =
            new G4PVPlacement(0,G4ThreeVector(0,0,fs_pos),absorber_cyl_log,"absorber",expHall_log,false,0);


    ///IRIS
    bool useiris=Conf()->GetBool("G4EusoGeometryShape.UseIRIS");
    if ( useiris ){
        double iris_r_inner=fOpticalSystemParam->surface[3].r_lens;
        double iris_r_outer=fOpticalSystemParam->r_wall;
        double iris_dz=fOpticalSystemParam->surface[3].Sz;
        //double iris_cut=fOpticalSystemParam->surface[3].r_cut;
        double iris_thickness=1.*mm;
        G4Tubs* iris_cyl=new G4Tubs("iris",iris_r_inner,iris_r_outer,iris_thickness,0.*deg,360.*deg);
	G4IntersectionSolid* intersec_iris = new G4IntersectionSolid("box*iris", box, iris_cyl);
        G4LogicalVolume* iris_log= new G4LogicalVolume(intersec_iris,fe_mat,"iris",0,0,0);

        /*G4VPhysicalVolume* iris_phys = */
        new G4PVPlacement(0,G4ThreeVector(0,0,iris_dz+iris_thickness),iris_log,"iris",expHall_log,false,0);
    }
    else Msg(EsafMsg::Warning) << "IRIS is turned off." << MsgDispatch;

    if(mp1)delete mp1;
    if(mp2)delete mp2;

// //     // Export VGM geometry to Root
//     Geant4GM::Factory g4Factory;
// //     g4Factory.SetDebug(1);
//     g4Factory.SetIgnore(1);
//     g4Factory.Import(expHall_phys);
//
//     RootGM::Factory rtFactory;
// //     rtFactory.SetDebug(1);
//     g4Factory.Export(&rtFactory);
//     gGeoManager->CloseGeometry();
//     gGeoManager->Export("output/G4lens.root");

    return expHall_phys;
}

void G4EusoGeometryShape::SetMaterial(int i, MaterialParameters* mp, char* material){

    char* matname = fOpticalSystemParam->material[i].name;
    const char* fmatname;
    if(!matname){
        matname = fOpticalSystemParam->material[i-1].name;
        fmatname = Form("%s/%s",lens_dir,fOpticalSystemParam->material[i-1].filename);
    }
    else fmatname = Form("%s/%s",lens_dir,fOpticalSystemParam->material[i].filename);
#ifdef DEBUG
    printf("%s %s %s %s\n",fOpticalSystemParam->material[0].name,fOpticalSystemParam->material[1].name,
                           fOpticalSystemParam->material[2].name,fOpticalSystemParam->material[3].name);
    printf("%s %s\n",fOpticalSystemParam->material[2].filename,fOpticalSystemParam->material[3].filename);
#endif


    if(!strcmp(matname,"CYTOP")){
        mp->setRI(fmatname);
        mp->SetData(CYTOP);
        mp->SetData(fOpticalSurfaceC);
        strcpy (material,"CYTOP");
    }
    else if(!strcmp(matname,"PMMA")){
           if(fit){
            TGraph* ph = new TGraph(fmatname);
            TF1* f=new TF1("ri","1./(x+[0])*[1]+1./(x*x+[2])*[3]+[4]",200.,600.);
            ph->Fit(f,"QNO","",250.,500.);//NOS
            mp->setRealRI(f,250.,500.);
            mp->setTrans(Form("%s/pmma_transmittivity.dat",lens_dir));
            delete ph;
            delete f;

        }
        else{
            mp->setRI(fmatname);
        }
        mp->SetData(PMMA);
        mp->SetData(fOpticalSurfaceP);
        strcpy (material,"PMMA");

    }
    else cerr<<"mat table: material not found"<<endl;


}
