#include "OpEUSODiffraction.hh"

#include <G4GeometryTolerance.hh>
#include <G4Step.hh>
#include <G4DynamicParticle.hh>
#include <G4Material.hh>
#include <G4OpticalPhoton.hh>
#include <G4TransportationManager.hh>
#include <G4Navigator.hh>
#include <G4Version.hh>

#include <TGraph.h>
#include "Ntrace_optics.hh"
using namespace NTraceLens;

//______________________________________________________________________________
OpEUSODiffraction::OpEUSODiffraction(const G4String& processName, G4ProcessType type) :
G4VDiscreteProcess(processName, type) {
        this->SetSender("OpEUSODiffraction");
        Msg(EsafMsg::Debug) << "OpEUSODiffraction is created." << MsgDispatch;

        kCarTolerance = G4GeometryTolerance::GetInstance()->GetSurfaceTolerance();
        const char* lensdir = Conf()->GetStr("G4EusoGeometryShape.lens_dir").data();
        char* name=fOpticalSystemParam->name;
        const char* dindex_fname = Form("%sDOE_%s.dat",lensdir,name);
        theDindex=new TGraph(dindex_fname);
        if ( !theDindex || theDindex->IsZombie() ){
            Msg(EsafMsg::Panic) << "File with DOE not found(" << dindex_fname << ")" << MsgDispatch;
        }
        double y;
        theDindex->GetPoint(theDindex->GetN()-1,theMaximumPerp,y);

        theLambda0=357.0; ///nm

}

//______________________________________________________________________________
OpEUSODiffraction::~OpEUSODiffraction(){
    delete theDindex;
}

//______________________________________________________________________________
G4bool OpEUSODiffraction::IsApplicable(const G4ParticleDefinition& aParticleType){
   return ( &aParticleType == G4OpticalPhoton::OpticalPhoton() );
}

//______________________________________________________________________________
G4VParticleChange*
OpEUSODiffraction::PostStepDoIt(const G4Track& aTrack, const G4Step& aStep)
{
    ///
    /// partly taken from the G4OpBoundaryProcess::PostStepDoIt() implementation
    ///
    aParticleChange.Initialize(aTrack);

    if (aTrack.GetStepLength()<=kCarTolerance*0.5){
            theStatus = edStepTooSmall;
            return G4VDiscreteProcess::PostStepDoIt(aTrack, aStep);
    }

    G4StepPoint* pPreStepPoint  = aStep.GetPreStepPoint();
    G4StepPoint* pPostStepPoint = aStep.GetPostStepPoint();

    if ( pPostStepPoint->GetStepStatus()!=fGeomBoundary ){
        theStatus = edNotAtBoundary;
        return G4VDiscreteProcess::PostStepDoIt(aTrack, aStep);
    }

    ///dynamic particle parameters
    const G4DynamicParticle* aParticle = aTrack.GetDynamicParticle();
    thePhotonMomentum = aParticle->GetTotalMomentum();
    OldMomentum       = aParticle->GetMomentumDirection();
    OldPolarization   = aParticle->GetPolarization();

    ///previous and next material parameters
    Material1 = pPreStepPoint->GetMaterial();
    Material2 = pPostStepPoint->GetMaterial();
    if ( Material1==Material2 ){
        theStatus = edSameMaterial;
        return G4VDiscreteProcess::PostStepDoIt(aTrack, aStep);
    }

    G4MaterialPropertyVector* Rindex;
    G4MaterialPropertiesTable* aMaterialPropertiesTable = Material1->GetMaterialPropertiesTable();
    if ( aMaterialPropertiesTable ) {
        Rindex = aMaterialPropertiesTable->GetProperty("EUSO_RINDEX");
    } else {
        theStatus = edNoINDEX;
        return G4VDiscreteProcess::PostStepDoIt(aTrack, aStep);
    }
    if ( !aMaterialPropertiesTable->GetProperty("EUSO_DINDEX") ){
        theStatus = edNoINDEX;
        return G4VDiscreteProcess::PostStepDoIt(aTrack, aStep);
    }
    if ( Rindex ) {
		#if G4VERSION_NUMBER<950
            Rindex1 = Rindex->GetProperty(thePhotonMomentum);//g9.4
		#else
        	Rindex1 = Rindex->Value(thePhotonMomentum); //g9.5
		#endif
    } else {
        theStatus = edNoINDEX;
        return G4VDiscreteProcess::PostStepDoIt(aTrack, aStep);
    }

    aMaterialPropertiesTable = Material2->GetMaterialPropertiesTable();
    if (aMaterialPropertiesTable) {
        Rindex = aMaterialPropertiesTable->GetProperty("RINDEX");
    } else {
        theStatus = edNoINDEX;
        return G4VDiscreteProcess::PostStepDoIt(aTrack, aStep);
    }

    if ( Rindex ) {
		#if G4VERSION_NUMBER<950
        	Rindex2 = Rindex->GetProperty(thePhotonMomentum);//g9.4
		#else
            Rindex2 = Rindex->Value(thePhotonMomentum);//g9.5
		#endif
    } else {
        theStatus = edNoINDEX;
        return G4VDiscreteProcess::PostStepDoIt(aTrack, aStep);
    }

    G4ThreeVector theGlobalPoint = pPostStepPoint->GetPosition();
    double perp=theGlobalPoint.perp();
    if ( perp > theMaximumPerp ){
        return G4VDiscreteProcess::PostStepDoIt(aTrack, aStep);
    }
    Dindex1 = theDindex->Eval( perp );

///adopted from Ltrace_optF1v4
/*
 * ------------------------------------------------------------------------
 *  calculate the diffracted ray
 *  Input:
 *        theGlobalPoint   array x,y,z of photon
 *        nnn[3]        normal vector to the diffractive surface at the intersection
 *        Rindex1            refractive index of xxx
 *        Rindex2            refractive index of xxx
 *        Dindex1
 *        wl            wavelength in mm
 *        m             index
 *  Output:
 *        in_ph[3]
 * ------------------------------------------------------------------------
 */
     G4ThreeVector nnnR(theGlobalPoint.x()/perp,theGlobalPoint.y()/perp,0.0);
//     G4ThreeVector nnn(.0,.0,1.0);

    G4Navigator* theNavigator = G4TransportationManager::GetTransportationManager()->GetNavigatorForTracking();
    G4ThreeVector nnn;   // Normal points back into volume
    G4bool valid;
    nnn = theNavigator->GetLocalExitNormal(&valid);
    if ( !valid ) {
        printf("Error: invalid normal\n");
        return G4VDiscreteProcess::PostStepDoIt(aTrack, aStep);
    }

    Rindex1=Rindex2=1.0003;                             // as in Ltrace, I don't know why
    double dir=OldMomentum.z()>0.?1.0:-1.0;
    double CosIN=OldMomentum*nnn;
    double CosD=nnn*nnnR;

    double m=1.;                                        // as in Ltrace
    double wl=hbarc*twopi/aTrack.GetTotalEnergy()*1.e6; // energy in MeV to lambda in nm
    double c1=Rindex1/Rindex2;
    double c2=wl/theLambda0*Dindex1*dir*m/Rindex2;
    G4ThreeVector ppp=(OldMomentum-nnn*CosIN)*c1+(nnnR-nnn*CosD)*c2;
    double norm=ppp.mag2();

    //#define DEBUG 1
    if(norm>1.0) {
        #ifdef DEBUG
              printf("norm>1, return\n");
        #endif
        return G4VDiscreteProcess::PostStepDoIt(aTrack, aStep);
    }
    norm=sqrt(1.0-norm);

    NewMomentum=ppp+nnn*norm*dir;
#ifdef DEBUG
    cout<<"our"<<endl;
    printf("(%.3f , %.3f , %.3f)\n", theGlobalPoint.x(),theGlobalPoint.y(),theGlobalPoint.z());
    printf("(%.3f , %.3f , %.3f)\n", nnnR.x(),nnnR.y(),nnnR.z());
    printf("(%.3f , %.3f , %.3f)\n", Rindex1,Dindex1,CosIN);
    printf("ppp: (%.3f , %.3f , %.3f)\n", ppp.x(),ppp.y(),ppp.z());
    printf("Diffractive: (%.3f , %.3f , %.3f)=>", OldMomentum.x(),OldMomentum.y(),OldMomentum.z());
    printf("(%.3f , %.3f , %.3f)\n", NewMomentum.x(),NewMomentum.y(),NewMomentum.z());
    printf("%g -> %g \n", OldMomentum.mag(), NewMomentum.mag());
#endif

    aParticleChange.ProposeMomentumDirection(NewMomentum);

    return G4VDiscreteProcess::PostStepDoIt(aTrack, aStep);

}


//______________________________________________________________________________
G4double OpEUSODiffraction::GetMeanFreePath(const G4Track& ,
                                              G4double ,
                                              G4ForceCondition* condition)
{
	*condition = Forced;
	return DBL_MAX;
}
