#include "OptUserStackingAction.hh"

#include "OpticsFactory.hh"
#include "IdealFocalSurface.hh"
#include "EDetectorPhotonAdder.hh"
#include "EVector.hh"
#include "EEvent.hh"
#include "EConst.hh"
#include "PhotonsOnPupil.hh"
#include "utils.hh"
#include "ConfigFileParser.hh"
#include "EsafConfigurable.hh"

#include <G4ParticleTable.hh>
#include <G4DynamicParticle.hh>
#include <G4Track.hh>
#include <Randomize.hh>
#include <G4StackManager.hh>
#include <G4RunManager.hh>
#include <G4Run.hh>
#include <G4TrackInformation.hh>
#include <G4VUserTrackInformation.hh>
#include <TArrayI.h>
#include <globals.hh>

#include "ParentPhoton.hh"

using namespace EConst;
// Detector subsystem identifier
// redefine to avoid kUndefined being defined twice
enum EDetectorSystem {
    keUndefined               = 0,
    kNotInside                = 1,
    kBaffle                   = 2,
    kOptics                   = 3,
    kFocalPlane               = 4,
    kWalls                    = 5,
    kEarth                    = 6,
    kIris                     = 7,
    kG4FocalSurface           = 8
};

void SetOptPhotonPolar(G4Track* track)
{
    G4double angle = G4UniformRand() * 360.0*deg;

    G4ThreeVector normal (1., 0., 0.);
    G4ThreeVector kphoton = track->GetMomentumDirection();
    G4ThreeVector product = normal.cross(kphoton);
    G4double modul2       = product*product;

    G4ThreeVector e_perpend (0., 0., 1.);
    if (modul2 > 0.) e_perpend = (1./std::sqrt(modul2))*product;
    G4ThreeVector e_paralle    = e_perpend.cross(kphoton);

    G4ThreeVector polar = std::cos(angle)*e_paralle + std::sin(angle)*e_perpend;
    track->SetPolarization(polar);
}

//_______________________________________________________________________________
OptUserStackingAction::OptUserStackingAction(IdealFocalSurface* fs) :
fPhotonDef(0),
fIdealFocalSurface(fs),
fHuge(1e20),
fNmax(-1),
fTrackHistory(0)
{
    fFocalPlane=OpticsFactory::Get()->GetFocalPlane();
    this->Reset();
    ConfigFileParser* conf = Config::Get()->GetCF("G4","G4Detector");
    string det=conf->GetStr("G4Detector.DetectorType");
    if (det.find("Tus")!=string::npos)ftus=true;
    else ftus=false;
		
    SetSender("OptStackingAction");
}

//_______________________________________________________________________________
OptUserStackingAction::~OptUserStackingAction(){
    // fFiller is deleted by Geant:
   // delete info;
   MsgForm(EsafMsg::Info,"fNTracked = %i.",fNTracked );
	 MsgForm(EsafMsg::Info,"walls = %i.",fNwalls );
	 MsgForm(EsafMsg::Info,"optics = %i.",fNoptics );
	 MsgForm(EsafMsg::Info,"focal plane = %i.",fNfocalplane );
	 MsgForm(EsafMsg::Info,"earth = %i.",fNearth );
   MsgForm(EsafMsg::Info,"iris = %i.",fNiris );
	 MsgForm(EsafMsg::Info,"lost = %i.",fNlost );
}


//_______________________________________________________________________________
void OptUserStackingAction::CheckTrace(int id,
                                           const G4ThreeVector& pos,
                                           const G4ThreeVector& mom,
                                           double lambda,
                                           double time)
{
    NTraceLens::NPhoton ph_in;
    ph_in.id=id;
    ph_in.x[0]=pos[0]/mm;
    ph_in.x[1]=pos[1]/mm;
    ph_in.x[2]=pos[2]/mm;
    ph_in.px[0]=mom[0];
    ph_in.px[1]=mom[1];
    ph_in.px[2]=mom[2];
    ph_in.id=id;
    ph_in.lambda=lambda;
    ph_in.t=time;
    ph_in.history=0;


#ifdef TRACEDEBUG
    cout<<"NOpticalSystem"<<endl;
    cout<<"position = "<<p.pos<<endl;
    cout<<"direction = "<<p.dir<<endl;
    double theta=p.dir.Angle(EVector(0.,0.,1.));
    cout<<"theta = "<<theta/deg<<endl;
    cout<<"wl = "<<p.wl/nm<<endl;
    cout<<"time = "<<p.time/ns<<endl;
#endif /* DEBUG */
}

//_______________________________________________________________________________
G4Track* OptUserStackingAction::NewPhoton(Photon* ph, int id, int refl){
    if ( !fPhotonDef ) {
        G4ParticleTable* particleTable = G4ParticleTable::GetParticleTable();
        fPhotonDef = particleTable->FindParticle("opticalphoton");
    }
    EVector& phdir=ph->dir;
    if(ftus)phdir*=-1.;
    double energy=hbarc*twopi/ph->wl; // lambda in mm to energy in MeV
    G4DynamicParticle* ph_dyn=new G4DynamicParticle(fPhotonDef,
                                                    G4ThreeVector(phdir[X],phdir[Y],phdir[Z]),
                                                    energy);
    EVector& phpos=ph->pos;
    /// system of ESAF is copied from CLHEP, so no change is needed
    G4ThreeVector pos(phpos[X],phpos[Y],phpos[Z]);
    G4ThreeVector mom(phdir[X],phdir[Y],phdir[Z]);
    double time;
    if(!ftus && !refl){
        double shift=100.*mm;
        pos-=mom.unit()*shift;
        time=ph->time-shift/c_light;
    }
    else time=ph->time;
    ph->time=time;
    G4Track* ph_tr=new G4Track( ph_dyn, time, pos);
    ph_tr->SetUserInformation(new G4TrackInformation(ph,refl,forigin));
    ph_tr->SetTrackID(id);
    SetOptPhotonPolar(ph_tr);

    if(refl)return ph_tr;
    if (fEv) {
        EDetectorPhotonAdder a(ph,0,kTRUE);
        fEv->Fill(a);
    }
    if ( ph->time > fTimeLastPhoton ) fTimeLastPhoton = ph->time;
    if ( ph->time < fTimeFirstPhoton ) fTimeFirstPhoton = ph->time;
    return ph_tr;
}

//_______________________________________________________________________________
Photon* OptUserStackingAction::UpdateCurrentPhoton(G4Track* track){
    G4Step* st = (G4Step*)track->GetStep();
    G4StepPoint *PostStep = st->GetPostStepPoint();
    G4ThreeVector pos =  PostStep->GetPosition();
    G4ThreeVector dir =  PostStep->GetMomentumDirection();
    double time = PostStep->GetLocalTime();
    G4TrackInformation* iph = (G4TrackInformation*)track->GetUserInformation();
    Photon* ph;
    double x[3] = {0.,0.,0.};
    double y[3] = {pos.x(),pos.y(),pos.z()};

    if(!iph){
        ph = new Photon();
        forigin=1;

        ph->time=0;
        ph->parent=0;
        ph->id=fNG4Tracked;
        double energy=track->GetKineticEnergy();
        double theConvConst=hbarc*twopi/eV;
        ph->wl=theConvConst/energy*nm;
        ParentPhoton *p = new ParentPhoton(fNG4Tracked,x, y, dir.theta(), 
                                           dir.phi(),theConvConst/energy,time,(ParentPhotonType)2);
        ph->parent=p;
        track->SetUserInformation(new G4TrackInformation(ph,0,1));
        fNG4Tracked++;

    }
    else ph = (Photon*)iph->fPhoton;
    ph->pos.SetXYZ(pos.x(),pos.y(),pos.z());
    ph->dir.SetXYZ(dir.x(),dir.y(),dir.z());
    ph->time+=time;
    ph->history=fTrackHistory;
    return ph;
}

//_______________________________________________________________________________
G4ClassificationOfNewTrack OptUserStackingAction::ClassifyNewTrack(const G4Track* aTrack){

//     if ( !fFiller ) {
//
//         return fKill;
//     }
//     if ( aTrack==fFiller ) {
//
//         return fWaiting;
//     }

    return fUrgent;
}

//______________________________________________________________________________
void OptUserStackingAction::NewStage(){

    // get a new photon
    if(!fNTotal)return;

    Photon* photon = fPupil->GetPhoton();

    // build an empty EDetPhoton in the event
    if ( !photon || ( fNmax>0 && fNTracked>=fNmax ) ) {


        if ( fEv ) {
            TArrayI photonHistory(8);
            photonHistory.AddAt(fNlost, 0);
            photonHistory.AddAt(fNG4focalplane, kG4FocalSurface);
            photonHistory.AddAt(fNfocalplane, kFocalPlane);
            photonHistory.AddAt(fNoptics, kOptics);

            EDetectorPhotonAdder a(photonHistory);
            fEv->Fill(a);
        }

        MsgForm(EsafMsg::Info,"Number of events processed = %i.",fNTracked );
        return;
    }

    forigin=0;
    stackManager->PushOneTrack( NewPhoton(photon, fNTracked,0) );
}


//_______________________________________________________________________________
void OptUserStackingAction::PrepareNewEvent(){

    this->NewStage();
}


//______________________________________________________________________________
void OptUserStackingAction::SetPupil(PhotonsOnPupil* p){
    printf("**********Set event\n");
    Reset();
    fEv = EEvent::GetCurrent();
    if(!p) return;

    
    fPupil=p;

    fNTotal=fPupil->GetNphotons();
    
}

//__________________________________________________________________
void OptUserStackingAction::Reset(){

    fNG4focalplane = 0;
    fNfocalplane = 0;
    fNlost = 0;
		fNearth = 0;
		fNoptics = 0;
		fNwalls = 0;
    fNTracked = 0;
    fNiris = 0;
    fNG4Tracked = 0;
    fProgress = 0;
    forigin = 0;
    
    fNTotal = 0;
    
    fEv = 0;
    
    fTimeFirstPhoton = fHuge;
    fTimeLastPhoton = -fHuge;
}

//__________________________________________________________________
EVector OptUserStackingAction::NewInteractionPoint(const Photon *p, double DZ) const {
    EVector dir=p->dir.Unit();
    EVector newPoint=p->pos+dir*(DZ/dir[Z]);
#ifdef TRACEDEBUG
    cout<<"new intPoint: "<<newPoint<<endl;
#endif /* DEBUG */
    return newPoint;
}


//-----------------------------------------------------------------
void OptUserStackingAction::PostUserTrackingAction(const G4Track* fTrackPost){

    if( fTrackPost->GetDefinition() == G4OpticalPhoton::OpticalPhotonDefinition()){
        Photon* ph = (Photon*)this->UpdateCurrentPhoton((G4Track*)fTrackPost);

        if ( this->GoToElectronics(ph) ){          
          return;
        }
        
        switch ( ph->history ) {
          case kFocalPlane:
            fNfocalplane++;
            break;
          case kG4FocalSurface :
            fNG4focalplane++;
            break;
          case kOptics:
            fNoptics++;
            break;
          case kWalls:
            fNwalls++;
            break;
          case kEarth:
            fNearth++;
            break;
          case kIris:
            fNiris++;
            break;
          default:
            fNlost++;
            printf("%i history *** x,y,z (%g %g %g)\n",ph->history, ph->pos[X],ph->pos[Y],ph->pos[Z]); 
        }
        
        if( fEv  ) {
//           bool is_new = kFALSE;
//           if(!fNTotal)is_new = kTRUE;
//           EDetectorPhotonAdder a(ph,0,is_new);
//           fEv->Fill(a);
        } 
        
    }
    
    
    if (fNTotal){
      fNTracked++;
      this->Progress(100.*fNTracked/fNTotal,"Photons");
    }
}
//______________________________________________________________________________
void OptUserStackingAction::Progress(int n,const char* p){
    int prog =n;

    if ( prog%10==0 && prog>=fProgress ) {
        Msg(EsafMsg::Info).SetProgress(fProgress);             //fProgress
        Msg(EsafMsg::Info) << Form("Tracking %s:  ",p) << MsgCount;
        fProgress+=10;

    }


}
//______________________________________________________________________________
bool OptUserStackingAction::GoToElectronics(Photon* ph){

    if ( ph->history!= kG4FocalSurface){
//         Msg(EsafMsg::Debug) << Form("photon isn't on FocalPlane %d",ph->history) << MsgDispatch;
//         Msg(EsafMsg::Debug) << "@ new photon " << MsgDispatch;
        return false;
    }
    EVector intPoint;
//     ph->history=kFocalPlane;

    if(ftus){
        ph->posOnIfs= ph->pos;
        ph->hitIfs=kTRUE;
    }
    else fIdealFocalSurface->HitPosition(ph);

    if(ph->hitIfs)ph->history=kFocalPlane;
      
    double DZ= fFocalPlane->Bottom()-ph->pos[Z];
    // calculate impact point on the focal plane
    intPoint=NewInteractionPoint(ph, DZ);
    ph->time+=(ph->pos - intPoint).Mag()/Clight();
    ph->pos=intPoint;
//     printf("1new!\n");
    
    Photon* newp = fFocalPlane->Transport(ph);
//     printf("0new!\n");
    if ( !newp ) return false;
//     printf("new!\n");
    stackManager->PushOneTrack( this->NewPhoton(newp,newp->id,1) );
    return true;
}

void OptUserStackingAction::EndOfEventAction(const G4Event* anEvent){
    int n=G4RunManager::GetRunManager()->GetCurrentRun()->GetNumberOfEventToBeProcessed();
    int curn=G4RunManager::GetRunManager()->GetCurrentRun()->GetNumberOfEvent();
    this->Progress(100.*curn/n,"events");

    if (  (curn+1)>=n ) {
        if ( fEv ) {
                TArrayI photonHistory(9);
                photonHistory.AddAt(fNlost, 0);
                photonHistory.AddAt(fNG4focalplane, kG4FocalSurface);
                photonHistory.AddAt(fNfocalplane, kFocalPlane);
                photonHistory.AddAt(fNoptics, kOptics);
                EDetectorPhotonAdder a(photonHistory);
                fEv->Fill(a);
        }
    }

}

void OptUserStackingAction::PreUserTrackingAction(const G4Track* fTrackPre){

}
void OptUserStackingAction::BeginOfEventAction(const G4Event* anEvent){
}
